"""                                                                                                      
Copyright 2016-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.                                  
                                                                                                         
Licensed under the Apache License, Version 2.0 (the "License"). You may not use this file except in compliance with the License. A copy of the License is located at                                              
                                                                                                         
    http://aws.amazon.com/apache2.0/                                                                     
                                                                                                         
or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.                                                 
"""

import json

def convert_join(sep, parts):
    """
    Fix a Join ;)
    """

    plain_string = True

    args = {}

    for i, part in enumerate(parts):
        if isinstance(part, dict):
            plain_string = False

            if "Ref" in part:
                parts[i] = "${{{}}}".format(part["Ref"])
            elif "Fn::GetAtt" in part:
                params = part["Fn::GetAtt"]
                parts[i] = "${{{}.{}}}".format(params[0], params[1])
            else:
                param_name = "Param{}".format(len(args) + 1)
                args[param_name] = part
                parts[i] = "${{{}}}".format(param_name)

    source = sep.join(parts)

    if plain_string:
        return source

    if args:
        return {
            "Fn::Sub": [source, args],
        }
    
    return {
        "Fn::Sub": source,
    }

def clean(source):
    """
    Clean up the source:
    * Replace use of Fn::Join with Fn::Sub
    """

    if isinstance(source, dict):
        for key, value in source.items():
            if key == "Fn::Join":
                return convert_join(value[0], value[1])
            elif key == "Fn::GetAtt":
                source[key] = "{}.{}".format(value[0], value[1])
            else:
                source[key] = clean(value)

    elif isinstance(source, list):
        return [clean(item) for item in source]

    return source



"""                                                                                                      
Copyright 2016-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.                                  
                                                                                                         
Licensed under the Apache License, Version 2.0 (the "License"). You may not use this file except in compliance with the License. A copy of the License is located at                                              
                                                                                                         
    http://aws.amazon.com/apache2.0/                                                                     
                                                                                                         
or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.                                                 
"""

import imp
import six
import collections

custom_yaml = imp.load_module("custom_yaml", *imp.find_module("yaml"))

TAG_MAP = "tag:yaml.org,2002:map"
TAG_STRING = "tag:yaml.org,2002:str"
UNCONVERTED_SUFFIXES = ["Ref", "Condition"]

def multi_constructor(loader, tag_suffix, node):
    """
    Deal with !Ref style function format
    """

    if tag_suffix not in UNCONVERTED_SUFFIXES:
        tag_suffix = "Fn::{}".format(tag_suffix)

    constructor = None

    if isinstance(node, custom_yaml.ScalarNode):
        constructor = loader.construct_scalar
    elif isinstance(node, custom_yaml.SequenceNode):
        constructor = loader.construct_sequence
    elif isinstance(node, custom_yaml.MappingNode):
        constructor = loader.construct_mapping
    else:
        raise "Bad tag: !{}".format(tag_suffix)

    return {
        tag_suffix: constructor(node)
    }

def construct_mapping(self, node, deep=False):
    """
    Use OrderedDict for maps
    """

    mapping = collections.OrderedDict()

    for key_node, value_node in node.value:
        key = self.construct_object(key_node, deep=deep)
        value = self.construct_object(value_node, deep=deep)

        mapping[key] = value

    return mapping

class odict_items(list):
    """
    Helper class to ensure ordering is preserved
    """

    def __init__(self, items):
        new_items = []

        for item in items:
            class C(type(item)):
                def __lt__(self, *args, **kwargs):
                    return False

            new_items.append(C(item))

        return super(odict_items, self).__init__(new_items)

    def sort(self):
        pass

class ODict(collections.OrderedDict):
    def __init__(self, *args, **kwargs):
        super(ODict, self).__init__(*args, **kwargs)

        items = odict_items(self.items())
        self.items = lambda: items

def representer(dumper, data):
    """
    Deal with !Ref style function format and OrderedDict
    """

    if len(data.keys()) != 1:
        data = ODict(data)

        return dumper.represent_mapping(TAG_MAP, data, flow_style=False)

    key = list(data)[0]
    tag = key

    if key not in UNCONVERTED_SUFFIXES and not key.startswith("Fn::"):
        return dumper.represent_mapping(TAG_MAP, data, flow_style=False)

    if key.startswith("Fn::"):
        tag = key[4:]

    tag = "!{}".format(tag)

    data = data[key]

    if isinstance(data, dict):
        return dumper.represent_mapping(tag, data, flow_style=False)
    elif isinstance(data, list):
        return dumper.represent_sequence(tag, data, flow_style=True)

    return dumper.represent_scalar(tag, data)

# Customise our yaml
custom_yaml.add_representer(six.text_type, lambda dumper, value: dumper.represent_scalar(TAG_STRING, value))
custom_yaml.add_multi_constructor("!", multi_constructor)
custom_yaml.add_constructor(TAG_MAP, construct_mapping)
custom_yaml.add_representer(collections.OrderedDict, representer)
custom_yaml.add_representer(dict, representer)



"""                                                                                                      
Copyright 2016-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.                                  
                                                                                                         
Licensed under the Apache License, Version 2.0 (the "License"). You may not use this file except in compliance with the License. A copy of the License is located at                                              
                                                                                                         
    http://aws.amazon.com/apache2.0/                                                                     
                                                                                                         
or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.                                                 
"""

from . import flip
import argparse
import sys

def main():
    """
    Figure out the input and output stream
    Then figure out the input format and set the opposing output format
    """

    # Set up the arg parser
    parser = argparse.ArgumentParser(description="AWS CloudFormation Template Flip is a tool that converts AWS CloudFormation templates between JSON and YAML formats, making use of the YAML format's short function syntax where possible.")
    parser.add_argument("-c", "--clean", action="store_true", help="Performs some opinionated cleanup on your template. For now, this just converts uses of Fn::Join to Fn::Sub.")
    parser.add_argument("input", nargs="?", type=argparse.FileType("r"), default=sys.stdin, help="File to read from. If you do not supply a file, input will be read from stdin.")
    parser.add_argument("output", nargs="?", type=argparse.FileType("w"), default=sys.stdout, help="File to write to. If you do not supply a file, output will be written to stdout.")

    args = parser.parse_args()

    template = args.input.read()

    try:
        args.output.write(flip(template, args.clean))
    except Exception as e:
        sys.stderr.write("{}\n".format(str(e)))
        sys.exit(1)



"""                                                                                                      
Copyright 2016-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.                                  
                                                                                                         
Licensed under the Apache License, Version 2.0 (the "License"). You may not use this file except in compliance with the License. A copy of the License is located at                                              
                                                                                                         
    http://aws.amazon.com/apache2.0/                                                                     
                                                                                                         
or in the "license" file accompanying this file. This file is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.                                                 
"""

from .clean import clean
from .custom_json import DateTimeAwareJsonEncoder
from .custom_yaml import custom_yaml
import collections
import json

class MyDumper(custom_yaml.Dumper):
  """
  Indent block sequences from parent using more common style
  ("  - entry"  vs "- entry").  
  Causes fewer problems with validation and tools.
  """
  
  def increase_indent(self,flow=False, indentless=False):
    return super(MyDumper,self).increase_indent(flow, False)

def to_json(template, clean_up=False):
    """
    Convert the data to json
    undoing yaml short syntax where detected
    """

    data = custom_yaml.load(template)

    if clean_up:
        data = clean(data)

    return json.dumps(data, indent=4, cls=DateTimeAwareJsonEncoder)

def to_yaml(template, clean_up=False):
    """
    Convert the data to yaml
    using yaml short syntax for functions where possible
    """

    data = json.loads(template, object_pairs_hook=collections.OrderedDict)

    if clean_up:
        data = clean(data)

    return custom_yaml.dump(data, Dumper=MyDumper, default_flow_style=False)

def flip(template, clean_up=False):
    """
    Figure out the input format and convert the data to the opposing output format
    """

    try:
        return to_yaml(template, clean_up)
    except ValueError:
        pass  # Hand over to the yaml parser

    try:
        return to_json(template, clean_up)
    except Exception as e:
        raise Exception("Could not determine the input format: {}", e)



from datetime import date, datetime, time
import json

class DateTimeAwareJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()

        return json.JSONEncoder.default(self, o)
